@@ -1,12 +1,15 @@
-from __future__ import absolute_import, division, print_function, with_statement
+from __future__ import absolute_import, division, print_function
 
+import gc
 import contextlib
 import datetime
 import functools
+import platform
 import sys
 import textwrap
 import time
 import weakref
+import warnings
 
 from tornado.concurrent import return_future, Future
 from tornado.escape import url_escape
@@ -15,7 +18,7 @@ from tornado.ioloop import IOLoop
 from tornado.log import app_log
 from tornado import stack_context
 from tornado.testing import AsyncHTTPTestCase, AsyncTestCase, ExpectLog, gen_test
-from tornado.test.util import unittest, skipOnTravis, skipBefore33, skipBefore35, skipNotCPython, exec_test
+from tornado.test.util import unittest, skipOnTravis, skipBefore33, skipBefore35, skipNotCPython, exec_test, ignore_deprecation  # noqa: E501
 from tornado.web import Application, RequestHandler, asynchronous, HTTPError
 
 from tornado import gen
@@ -25,12 +28,24 @@ try:
 except ImportError:
     futures = None
 
+try:
+    import asyncio
+except ImportError:
+    asyncio = None
+
 
 class GenEngineTest(AsyncTestCase):
     def setUp(self):
+        self.warning_catcher = warnings.catch_warnings()
+        self.warning_catcher.__enter__()
+        warnings.simplefilter('ignore', DeprecationWarning)
         super(GenEngineTest, self).setUp()
         self.named_contexts = []
 
+    def tearDown(self):
+        super(GenEngineTest, self).tearDown()
+        self.warning_catcher.__exit__(None, None, None)
+
     def named_context(self, name):
         @contextlib.contextmanager
         def context():
@@ -53,9 +68,10 @@ class GenEngineTest(AsyncTestCase):
             self.io_loop.add_callback(functools.partial(
                 self.delay_callback, iterations - 1, callback, arg))
 
-    @return_future
-    def async_future(self, result, callback):
-        self.io_loop.add_callback(callback, result)
+    with ignore_deprecation():
+        @return_future
+        def async_future(self, result, callback):
+            self.io_loop.add_callback(callback, result)
 
     @gen.coroutine
     def async_exception(self, e):
@@ -276,6 +292,13 @@ class GenEngineTest(AsyncTestCase):
             pass
         self.orphaned_callback()
 
+    def test_none(self):
+        @gen.engine
+        def f():
+            yield None
+            self.stop()
+        self.run_gen(f)
+
     def test_multi(self):
         @gen.engine
         def f():
@@ -645,6 +668,237 @@ class GenEngineTest(AsyncTestCase):
         self.assertIs(self.task_ref(), None)
 
 
+# GenBasicTest duplicates the non-deprecated portions of GenEngineTest
+# with gen.coroutine to ensure we don't lose coverage when gen.engine
+# goes away.
+class GenBasicTest(AsyncTestCase):
+    @gen.coroutine
+    def delay(self, iterations, arg):
+        """Returns arg after a number of IOLoop iterations."""
+        for i in range(iterations):
+            yield gen.moment
+        raise gen.Return(arg)
+
+    with ignore_deprecation():
+        @return_future
+        def async_future(self, result, callback):
+            self.io_loop.add_callback(callback, result)
+
+    @gen.coroutine
+    def async_exception(self, e):
+        yield gen.moment
+        raise e
+
+    @gen.coroutine
+    def add_one_async(self, x):
+        yield gen.moment
+        raise gen.Return(x + 1)
+
+    def test_no_yield(self):
+        @gen.coroutine
+        def f():
+            pass
+        self.io_loop.run_sync(f)
+
+    def test_exception_phase1(self):
+        @gen.coroutine
+        def f():
+            1 / 0
+        self.assertRaises(ZeroDivisionError, self.io_loop.run_sync, f)
+
+    def test_exception_phase2(self):
+        @gen.coroutine
+        def f():
+            yield gen.moment
+            1 / 0
+        self.assertRaises(ZeroDivisionError, self.io_loop.run_sync, f)
+
+    def test_bogus_yield(self):
+        @gen.coroutine
+        def f():
+            yield 42
+        self.assertRaises(gen.BadYieldError, self.io_loop.run_sync, f)
+
+    def test_bogus_yield_tuple(self):
+        @gen.coroutine
+        def f():
+            yield (1, 2)
+        self.assertRaises(gen.BadYieldError, self.io_loop.run_sync, f)
+
+    def test_reuse(self):
+        @gen.coroutine
+        def f():
+            yield gen.moment
+        self.io_loop.run_sync(f)
+        self.io_loop.run_sync(f)
+
+    def test_none(self):
+        @gen.coroutine
+        def f():
+            yield None
+        self.io_loop.run_sync(f)
+
+    def test_multi(self):
+        @gen.coroutine
+        def f():
+            results = yield [self.add_one_async(1), self.add_one_async(2)]
+            self.assertEqual(results, [2, 3])
+        self.io_loop.run_sync(f)
+
+    def test_multi_dict(self):
+        @gen.coroutine
+        def f():
+            results = yield dict(foo=self.add_one_async(1), bar=self.add_one_async(2))
+            self.assertEqual(results, dict(foo=2, bar=3))
+        self.io_loop.run_sync(f)
+
+    def test_multi_delayed(self):
+        @gen.coroutine
+        def f():
+            # callbacks run at different times
+            responses = yield gen.multi_future([
+                self.delay(3, "v1"),
+                self.delay(1, "v2"),
+            ])
+            self.assertEqual(responses, ["v1", "v2"])
+        self.io_loop.run_sync(f)
+
+    def test_multi_dict_delayed(self):
+        @gen.coroutine
+        def f():
+            # callbacks run at different times
+            responses = yield gen.multi_future(dict(
+                foo=self.delay(3, "v1"),
+                bar=self.delay(1, "v2"),
+            ))
+            self.assertEqual(responses, dict(foo="v1", bar="v2"))
+        self.io_loop.run_sync(f)
+
+    @skipOnTravis
+    @gen_test
+    def test_multi_performance(self):
+        # Yielding a list used to have quadratic performance; make
+        # sure a large list stays reasonable.  On my laptop a list of
+        # 2000 used to take 1.8s, now it takes 0.12.
+        start = time.time()
+        yield [gen.moment for i in range(2000)]
+        end = time.time()
+        self.assertLess(end - start, 1.0)
+
+    @gen_test
+    def test_multi_empty(self):
+        # Empty lists or dicts should return the same type.
+        x = yield []
+        self.assertTrue(isinstance(x, list))
+        y = yield {}
+        self.assertTrue(isinstance(y, dict))
+
+    @gen_test
+    def test_future(self):
+        result = yield self.async_future(1)
+        self.assertEqual(result, 1)
+
+    @gen_test
+    def test_multi_future(self):
+        results = yield [self.async_future(1), self.async_future(2)]
+        self.assertEqual(results, [1, 2])
+
+    @gen_test
+    def test_multi_future_duplicate(self):
+        f = self.async_future(2)
+        results = yield [self.async_future(1), f, self.async_future(3), f]
+        self.assertEqual(results, [1, 2, 3, 2])
+
+    @gen_test
+    def test_multi_dict_future(self):
+        results = yield dict(foo=self.async_future(1), bar=self.async_future(2))
+        self.assertEqual(results, dict(foo=1, bar=2))
+
+    @gen_test
+    def test_multi_exceptions(self):
+        with ExpectLog(app_log, "Multiple exceptions in yield list"):
+            with self.assertRaises(RuntimeError) as cm:
+                yield gen.Multi([self.async_exception(RuntimeError("error 1")),
+                                 self.async_exception(RuntimeError("error 2"))])
+        self.assertEqual(str(cm.exception), "error 1")
+
+        # With only one exception, no error is logged.
+        with self.assertRaises(RuntimeError):
+            yield gen.Multi([self.async_exception(RuntimeError("error 1")),
+                             self.async_future(2)])
+
+        # Exception logging may be explicitly quieted.
+        with self.assertRaises(RuntimeError):
+            yield gen.Multi([self.async_exception(RuntimeError("error 1")),
+                             self.async_exception(RuntimeError("error 2"))],
+                            quiet_exceptions=RuntimeError)
+
+    @gen_test
+    def test_multi_future_exceptions(self):
+        with ExpectLog(app_log, "Multiple exceptions in yield list"):
+            with self.assertRaises(RuntimeError) as cm:
+                yield [self.async_exception(RuntimeError("error 1")),
+                       self.async_exception(RuntimeError("error 2"))]
+        self.assertEqual(str(cm.exception), "error 1")
+
+        # With only one exception, no error is logged.
+        with self.assertRaises(RuntimeError):
+            yield [self.async_exception(RuntimeError("error 1")),
+                   self.async_future(2)]
+
+        # Exception logging may be explicitly quieted.
+        with self.assertRaises(RuntimeError):
+            yield gen.multi_future(
+                [self.async_exception(RuntimeError("error 1")),
+                 self.async_exception(RuntimeError("error 2"))],
+                quiet_exceptions=RuntimeError)
+
+    def test_sync_raise_return(self):
+        @gen.coroutine
+        def f():
+            raise gen.Return()
+
+        self.io_loop.run_sync(f)
+
+    def test_async_raise_return(self):
+        @gen.coroutine
+        def f():
+            yield gen.moment
+            raise gen.Return()
+
+        self.io_loop.run_sync(f)
+
+    def test_sync_raise_return_value(self):
+        @gen.coroutine
+        def f():
+            raise gen.Return(42)
+
+        self.assertEqual(42, self.io_loop.run_sync(f))
+
+    def test_sync_raise_return_value_tuple(self):
+        @gen.coroutine
+        def f():
+            raise gen.Return((1, 2))
+
+        self.assertEqual((1, 2), self.io_loop.run_sync(f))
+
+    def test_async_raise_return_value(self):
+        @gen.coroutine
+        def f():
+            yield gen.moment
+            raise gen.Return(42)
+
+        self.assertEqual(42, self.io_loop.run_sync(f))
+
+    def test_async_raise_return_value_tuple(self):
+        @gen.coroutine
+        def f():
+            yield gen.moment
+            raise gen.Return((1, 2))
+
+        self.assertEqual((1, 2), self.io_loop.run_sync(f))
+
+
 class GenCoroutineTest(AsyncTestCase):
     def setUp(self):
         # Stray StopIteration exceptions can lead to tests exiting prematurely,
@@ -657,6 +911,28 @@ class GenCoroutineTest(AsyncTestCase):
         super(GenCoroutineTest, self).tearDown()
         assert self.finished
 
+    def test_attributes(self):
+        self.finished = True
+
+        def f():
+            yield gen.moment
+
+        coro = gen.coroutine(f)
+        self.assertEqual(coro.__name__, f.__name__)
+        self.assertEqual(coro.__module__, f.__module__)
+        self.assertIs(coro.__wrapped__, f)
+
+    def test_is_coroutine_function(self):
+        self.finished = True
+
+        def f():
+            yield gen.moment
+
+        coro = gen.coroutine(f)
+        self.assertFalse(gen.is_coroutine_function(f))
+        self.assertTrue(gen.is_coroutine_function(coro))
+        self.assertFalse(gen.is_coroutine_function(coro()))
+
     @gen_test
     def test_sync_gen_return(self):
         @gen.coroutine
@@ -670,7 +946,7 @@ class GenCoroutineTest(AsyncTestCase):
     def test_async_gen_return(self):
         @gen.coroutine
         def f():
-            yield gen.Task(self.io_loop.add_callback)
+            yield gen.moment
             raise gen.Return(42)
         result = yield f()
         self.assertEqual(result, 42)
@@ -691,7 +967,7 @@ class GenCoroutineTest(AsyncTestCase):
         namespace = exec_test(globals(), locals(), """
         @gen.coroutine
         def f():
-            yield gen.Task(self.io_loop.add_callback)
+            yield gen.moment
             return 42
         """)
         result = yield namespace['f']()
@@ -718,12 +994,32 @@ class GenCoroutineTest(AsyncTestCase):
     @skipBefore35
     @gen_test
     def test_async_await(self):
+        @gen.coroutine
+        def f1():
+            yield gen.moment
+            raise gen.Return(42)
+
         # This test verifies that an async function can await a
         # yield-based gen.coroutine, and that a gen.coroutine
         # (the test method itself) can yield an async function.
         namespace = exec_test(globals(), locals(), """
+        async def f2():
+            result = await f1()
+            return result
+        """)
+        result = yield namespace['f2']()
+        self.assertEqual(result, 42)
+        self.finished = True
+
+    @skipBefore35
+    @gen_test
+    def test_asyncio_sleep_zero(self):
+        # asyncio.sleep(0) turns into a special case (equivalent to
+        # `yield None`)
+        namespace = exec_test(globals(), locals(), """
         async def f():
-            await gen.Task(self.io_loop.add_callback)
+            import asyncio
+            await asyncio.sleep(0)
             return 42
         """)
         result = yield namespace['f']()
@@ -733,18 +1029,22 @@ class GenCoroutineTest(AsyncTestCase):
     @skipBefore35
     @gen_test
     def test_async_await_mixed_multi_native_future(self):
+        @gen.coroutine
+        def f1():
+            yield gen.moment
+
         namespace = exec_test(globals(), locals(), """
-        async def f1():
-            await gen.Task(self.io_loop.add_callback)
+        async def f2():
+            await f1()
             return 42
         """)
 
         @gen.coroutine
-        def f2():
-            yield gen.Task(self.io_loop.add_callback)
+        def f3():
+            yield gen.moment
             raise gen.Return(43)
 
-        results = yield [namespace['f1'](), f2()]
+        results = yield [namespace['f2'](), f3()]
         self.assertEqual(results, [42, 43])
         self.finished = True
 
@@ -762,8 +1062,9 @@ class GenCoroutineTest(AsyncTestCase):
             yield gen.Task(self.io_loop.add_callback)
             raise gen.Return(43)
 
-        f2(callback=(yield gen.Callback('cb')))
-        results = yield [namespace['f1'](), gen.Wait('cb')]
+        with ignore_deprecation():
+            f2(callback=(yield gen.Callback('cb')))
+            results = yield [namespace['f1'](), gen.Wait('cb')]
         self.assertEqual(results, [42, 43])
         self.finished = True
 
@@ -794,7 +1095,7 @@ class GenCoroutineTest(AsyncTestCase):
         # Without a return value we don't need python 3.3.
         @gen.coroutine
         def f():
-            yield gen.Task(self.io_loop.add_callback)
+            yield gen.moment
             return
         result = yield f()
         self.assertEqual(result, None)
@@ -817,7 +1118,7 @@ class GenCoroutineTest(AsyncTestCase):
     def test_async_raise(self):
         @gen.coroutine
         def f():
-            yield gen.Task(self.io_loop.add_callback)
+            yield gen.moment
             1 / 0
         future = f()
         with self.assertRaises(ZeroDivisionError):
@@ -826,10 +1127,11 @@ class GenCoroutineTest(AsyncTestCase):
 
     @gen_test
     def test_pass_callback(self):
-        @gen.coroutine
-        def f():
-            raise gen.Return(42)
-        result = yield gen.Task(f)
+        with ignore_deprecation():
+            @gen.coroutine
+            def f():
+                raise gen.Return(42)
+            result = yield gen.Task(f)
         self.assertEqual(result, 42)
         self.finished = True
 
@@ -874,46 +1176,48 @@ class GenCoroutineTest(AsyncTestCase):
 
     @gen_test
     def test_replace_context_exception(self):
-        # Test exception handling: exceptions thrown into the stack context
-        # can be caught and replaced.
-        # Note that this test and the following are for behavior that is
-        # not really supported any more:  coroutines no longer create a
-        # stack context automatically; but one is created after the first
-        # YieldPoint (i.e. not a Future).
-        @gen.coroutine
-        def f2():
-            (yield gen.Callback(1))()
-            yield gen.Wait(1)
-            self.io_loop.add_callback(lambda: 1 / 0)
-            try:
-                yield gen.Task(self.io_loop.add_timeout,
-                               self.io_loop.time() + 10)
-            except ZeroDivisionError:
-                raise KeyError()
+        with ignore_deprecation():
+            # Test exception handling: exceptions thrown into the stack context
+            # can be caught and replaced.
+            # Note that this test and the following are for behavior that is
+            # not really supported any more:  coroutines no longer create a
+            # stack context automatically; but one is created after the first
+            # YieldPoint (i.e. not a Future).
+            @gen.coroutine
+            def f2():
+                (yield gen.Callback(1))()
+                yield gen.Wait(1)
+                self.io_loop.add_callback(lambda: 1 / 0)
+                try:
+                    yield gen.Task(self.io_loop.add_timeout,
+                                   self.io_loop.time() + 10)
+                except ZeroDivisionError:
+                    raise KeyError()
 
-        future = f2()
-        with self.assertRaises(KeyError):
-            yield future
-        self.finished = True
+            future = f2()
+            with self.assertRaises(KeyError):
+                yield future
+            self.finished = True
 
     @gen_test
     def test_swallow_context_exception(self):
-        # Test exception handling: exceptions thrown into the stack context
-        # can be caught and ignored.
-        @gen.coroutine
-        def f2():
-            (yield gen.Callback(1))()
-            yield gen.Wait(1)
-            self.io_loop.add_callback(lambda: 1 / 0)
-            try:
-                yield gen.Task(self.io_loop.add_timeout,
-                               self.io_loop.time() + 10)
-            except ZeroDivisionError:
-                raise gen.Return(42)
+        with ignore_deprecation():
+            # Test exception handling: exceptions thrown into the stack context
+            # can be caught and ignored.
+            @gen.coroutine
+            def f2():
+                (yield gen.Callback(1))()
+                yield gen.Wait(1)
+                self.io_loop.add_callback(lambda: 1 / 0)
+                try:
+                    yield gen.Task(self.io_loop.add_timeout,
+                                   self.io_loop.time() + 10)
+                except ZeroDivisionError:
+                    raise gen.Return(42)
 
-        result = yield f2()
-        self.assertEqual(result, 42)
-        self.finished = True
+            result = yield f2()
+            self.assertEqual(result, 42)
+            self.finished = True
 
     @gen_test
     def test_moment(self):
@@ -970,77 +1274,132 @@ class GenCoroutineTest(AsyncTestCase):
 
         self.finished = True
 
+    @skipNotCPython
+    @unittest.skipIf((3,) < sys.version_info < (3, 6),
+                     "asyncio.Future has reference cycles")
+    def test_coroutine_refcounting(self):
+        # On CPython, tasks and their arguments should be released immediately
+        # without waiting for garbage collection.
+        @gen.coroutine
+        def inner():
+            class Foo(object):
+                pass
+            local_var = Foo()
+            self.local_ref = weakref.ref(local_var)
+            yield gen.coroutine(lambda: None)()
+            raise ValueError('Some error')
+
+        @gen.coroutine
+        def inner2():
+            try:
+                yield inner()
+            except ValueError:
+                pass
+
+        self.io_loop.run_sync(inner2, timeout=3)
+
+        self.assertIs(self.local_ref(), None)
+        self.finished = True
+
+    @unittest.skipIf(sys.version_info < (3,),
+                     "test only relevant with asyncio Futures")
+    def test_asyncio_future_debug_info(self):
+        self.finished = True
+        # Enable debug mode
+        asyncio_loop = asyncio.get_event_loop()
+        self.addCleanup(asyncio_loop.set_debug, asyncio_loop.get_debug())
+        asyncio_loop.set_debug(True)
+
+        def f():
+            yield gen.moment
+
+        coro = gen.coroutine(f)()
+        self.assertIsInstance(coro, asyncio.Future)
+        # We expect the coroutine repr() to show the place where
+        # it was instantiated
+        expected = ("created at %s:%d"
+                    % (__file__, f.__code__.co_firstlineno + 3))
+        actual = repr(coro)
+        self.assertIn(expected, actual)
+
+    @unittest.skipIf(asyncio is None, "asyncio module not present")
+    @gen_test
+    def test_asyncio_gather(self):
+        # This demonstrates that tornado coroutines can be understood
+        # by asyncio (This failed prior to Tornado 5.0).
+        @gen.coroutine
+        def f():
+            yield gen.moment
+            raise gen.Return(1)
+
+        ret = yield asyncio.gather(f(), f())
+        self.assertEqual(ret, [1, 1])
+        self.finished = True
+
 
 class GenSequenceHandler(RequestHandler):
-    @asynchronous
-    @gen.engine
-    def get(self):
-        self.io_loop = self.request.connection.stream.io_loop
-        self.io_loop.add_callback((yield gen.Callback("k1")))
-        yield gen.Wait("k1")
-        self.write("1")
-        self.io_loop.add_callback((yield gen.Callback("k2")))
-        yield gen.Wait("k2")
-        self.write("2")
-        # reuse an old key
-        self.io_loop.add_callback((yield gen.Callback("k1")))
-        yield gen.Wait("k1")
-        self.finish("3")
+    with ignore_deprecation():
+        @asynchronous
+        @gen.engine
+        def get(self):
+            # The outer ignore_deprecation applies at definition time.
+            # We need another for serving time.
+            with ignore_deprecation():
+                self.io_loop = self.request.connection.stream.io_loop
+                self.io_loop.add_callback((yield gen.Callback("k1")))
+                yield gen.Wait("k1")
+                self.write("1")
+                self.io_loop.add_callback((yield gen.Callback("k2")))
+                yield gen.Wait("k2")
+                self.write("2")
+                # reuse an old key
+                self.io_loop.add_callback((yield gen.Callback("k1")))
+                yield gen.Wait("k1")
+                self.finish("3")
 
 
 class GenCoroutineSequenceHandler(RequestHandler):
     @gen.coroutine
     def get(self):
-        self.io_loop = self.request.connection.stream.io_loop
-        self.io_loop.add_callback((yield gen.Callback("k1")))
-        yield gen.Wait("k1")
+        yield gen.moment
         self.write("1")
-        self.io_loop.add_callback((yield gen.Callback("k2")))
-        yield gen.Wait("k2")
+        yield gen.moment
         self.write("2")
-        # reuse an old key
-        self.io_loop.add_callback((yield gen.Callback("k1")))
-        yield gen.Wait("k1")
+        yield gen.moment
         self.finish("3")
 
 
 class GenCoroutineUnfinishedSequenceHandler(RequestHandler):
-    @asynchronous
     @gen.coroutine
     def get(self):
-        self.io_loop = self.request.connection.stream.io_loop
-        self.io_loop.add_callback((yield gen.Callback("k1")))
-        yield gen.Wait("k1")
+        yield gen.moment
         self.write("1")
-        self.io_loop.add_callback((yield gen.Callback("k2")))
-        yield gen.Wait("k2")
+        yield gen.moment
         self.write("2")
-        # reuse an old key
-        self.io_loop.add_callback((yield gen.Callback("k1")))
-        yield gen.Wait("k1")
+        yield gen.moment
         # just write, don't finish
         self.write("3")
 
 
 class GenTaskHandler(RequestHandler):
-    @asynchronous
-    @gen.engine
+    @gen.coroutine
     def get(self):
-        io_loop = self.request.connection.stream.io_loop
-        client = AsyncHTTPClient(io_loop=io_loop)
-        response = yield gen.Task(client.fetch, self.get_argument('url'))
+        client = AsyncHTTPClient()
+        with ignore_deprecation():
+            response = yield gen.Task(client.fetch, self.get_argument('url'))
         response.rethrow()
         self.finish(b"got response: " + response.body)
 
 
 class GenExceptionHandler(RequestHandler):
-    @asynchronous
-    @gen.engine
-    def get(self):
-        # This test depends on the order of the two decorators.
-        io_loop = self.request.connection.stream.io_loop
-        yield gen.Task(io_loop.add_callback)
-        raise Exception("oops")
+    with ignore_deprecation():
+        @asynchronous
+        @gen.engine
+        def get(self):
+            # This test depends on the order of the two decorators.
+            io_loop = self.request.connection.stream.io_loop
+            yield gen.Task(io_loop.add_callback)
+            raise Exception("oops")
 
 
 class GenCoroutineExceptionHandler(RequestHandler):
@@ -1053,19 +1412,18 @@ class GenCoroutineExceptionHandler(RequestHandler):
 
 
 class GenYieldExceptionHandler(RequestHandler):
-    @asynchronous
-    @gen.engine
+    @gen.coroutine
     def get(self):
         io_loop = self.request.connection.stream.io_loop
         # Test the interaction of the two stack_contexts.
-
-        def fail_task(callback):
-            io_loop.add_callback(lambda: 1 / 0)
-        try:
-            yield gen.Task(fail_task)
-            raise Exception("did not get expected exception")
-        except ZeroDivisionError:
-            self.finish('ok')
+        with ignore_deprecation():
+            def fail_task(callback):
+                io_loop.add_callback(lambda: 1 / 0)
+            try:
+                yield gen.Task(fail_task)
+                raise Exception("did not get expected exception")
+            except ZeroDivisionError:
+                self.finish('ok')
 
 
 # "Undecorated" here refers to the absence of @asynchronous.
@@ -1073,22 +1431,22 @@ class UndecoratedCoroutinesHandler(RequestHandler):
     @gen.coroutine
     def prepare(self):
         self.chunks = []
-        yield gen.Task(IOLoop.current().add_callback)
+        yield gen.moment
         self.chunks.append('1')
 
     @gen.coroutine
     def get(self):
         self.chunks.append('2')
-        yield gen.Task(IOLoop.current().add_callback)
+        yield gen.moment
         self.chunks.append('3')
-        yield gen.Task(IOLoop.current().add_callback)
+        yield gen.moment
         self.write(''.join(self.chunks))
 
 
 class AsyncPrepareErrorHandler(RequestHandler):
     @gen.coroutine
     def prepare(self):
-        yield gen.Task(IOLoop.current().add_callback)
+        yield gen.moment
         raise HTTPError(403)
 
     def get(self):
@@ -1099,7 +1457,8 @@ class NativeCoroutineHandler(RequestHandler):
     if sys.version_info > (3, 5):
         exec(textwrap.dedent("""
         async def get(self):
-            await gen.Task(IOLoop.current().add_callback)
+            import asyncio
+            await asyncio.sleep(0)
             self.write("ok")
         """))
 
@@ -1180,7 +1539,7 @@ class WithTimeoutTest(AsyncTestCase):
         self.io_loop.add_timeout(datetime.timedelta(seconds=0.1),
                                  lambda: future.set_result('asdf'))
         result = yield gen.with_timeout(datetime.timedelta(seconds=3600),
-                                        future, io_loop=self.io_loop)
+                                        future)
         self.assertEqual(result, 'asdf')
 
     @gen_test
@@ -1191,19 +1550,20 @@ class WithTimeoutTest(AsyncTestCase):
             lambda: future.set_exception(ZeroDivisionError()))
         with self.assertRaises(ZeroDivisionError):
             yield gen.with_timeout(datetime.timedelta(seconds=3600),
-                                   future, io_loop=self.io_loop)
+                                   future)
 
     @gen_test
     def test_already_resolved(self):
         future = Future()
         future.set_result('asdf')
         result = yield gen.with_timeout(datetime.timedelta(seconds=3600),
-                                        future, io_loop=self.io_loop)
+                                        future)
         self.assertEqual(result, 'asdf')
 
     @unittest.skipIf(futures is None, 'futures module not present')
     @gen_test
     def test_timeout_concurrent_future(self):
+        # A concurrent future that does not resolve before the timeout.
         with futures.ThreadPoolExecutor(1) as executor:
             with self.assertRaises(gen.TimeoutError):
                 yield gen.with_timeout(self.io_loop.time(),
@@ -1212,9 +1572,20 @@ class WithTimeoutTest(AsyncTestCase):
     @unittest.skipIf(futures is None, 'futures module not present')
     @gen_test
     def test_completed_concurrent_future(self):
+        # A concurrent future that is resolved before we even submit it
+        # to with_timeout.
+        with futures.ThreadPoolExecutor(1) as executor:
+            f = executor.submit(lambda: None)
+            f.result()  # wait for completion
+            yield gen.with_timeout(datetime.timedelta(seconds=3600), f)
+
+    @unittest.skipIf(futures is None, 'futures module not present')
+    @gen_test
+    def test_normal_concurrent_future(self):
+        # A conccurrent future that resolves while waiting for the timeout.
         with futures.ThreadPoolExecutor(1) as executor:
             yield gen.with_timeout(datetime.timedelta(seconds=3600),
-                                   executor.submit(lambda: None))
+                                   executor.submit(lambda: time.sleep(0.01)))
 
 
 class WaitIteratorTest(AsyncTestCase):
@@ -1368,5 +1739,124 @@ class WaitIteratorTest(AsyncTestCase):
                                gen.WaitIterator(gen.sleep(0)).next())
 
 
+class RunnerGCTest(AsyncTestCase):
+    def is_pypy3(self):
+        return (platform.python_implementation() == 'PyPy' and
+                sys.version_info > (3,))
+
+    @gen_test
+    def test_gc(self):
+        # Github issue 1769: Runner objects can get GCed unexpectedly
+        # while their future is alive.
+        weakref_scope = [None]
+
+        def callback():
+            gc.collect(2)
+            weakref_scope[0]().set_result(123)
+
+        @gen.coroutine
+        def tester():
+            fut = Future()
+            weakref_scope[0] = weakref.ref(fut)
+            self.io_loop.add_callback(callback)
+            yield fut
+
+        yield gen.with_timeout(
+            datetime.timedelta(seconds=0.2),
+            tester()
+        )
+
+    def test_gc_infinite_coro(self):
+        # Github issue 2229: suspended coroutines should be GCed when
+        # their loop is closed, even if they're involved in a reference
+        # cycle.
+        if IOLoop.configured_class().__name__.endswith('TwistedIOLoop'):
+            raise unittest.SkipTest("Test may fail on TwistedIOLoop")
+
+        loop = self.get_new_ioloop()
+        result = []
+        wfut = []
+
+        @gen.coroutine
+        def infinite_coro():
+            try:
+                while True:
+                    yield gen.sleep(1e-3)
+                    result.append(True)
+            finally:
+                # coroutine finalizer
+                result.append(None)
+
+        @gen.coroutine
+        def do_something():
+            fut = infinite_coro()
+            fut._refcycle = fut
+            wfut.append(weakref.ref(fut))
+            yield gen.sleep(0.2)
+
+        loop.run_sync(do_something)
+        loop.close()
+        gc.collect()
+        # Future was collected
+        self.assertIs(wfut[0](), None)
+        # At least one wakeup
+        self.assertGreaterEqual(len(result), 2)
+        if not self.is_pypy3():
+            # coroutine finalizer was called (not on PyPy3 apparently)
+            self.assertIs(result[-1], None)
+
+    @skipBefore35
+    def test_gc_infinite_async_await(self):
+        # Same as test_gc_infinite_coro, but with a `async def` function
+        import asyncio
+
+        namespace = exec_test(globals(), locals(), """
+        async def infinite_coro(result):
+            try:
+                while True:
+                    await gen.sleep(1e-3)
+                    result.append(True)
+            finally:
+                # coroutine finalizer
+                result.append(None)
+        """)
+
+        infinite_coro = namespace['infinite_coro']
+        loop = self.get_new_ioloop()
+        result = []
+        wfut = []
+
+        @gen.coroutine
+        def do_something():
+            fut = asyncio.get_event_loop().create_task(infinite_coro(result))
+            fut._refcycle = fut
+            wfut.append(weakref.ref(fut))
+            yield gen.sleep(0.2)
+
+        loop.run_sync(do_something)
+        with ExpectLog('asyncio', "Task was destroyed but it is pending"):
+            loop.close()
+            gc.collect()
+        # Future was collected
+        self.assertIs(wfut[0](), None)
+        # At least one wakeup and one finally
+        self.assertGreaterEqual(len(result), 2)
+        if not self.is_pypy3():
+            # coroutine finalizer was called (not on PyPy3 apparently)
+            self.assertIs(result[-1], None)
+
+    def test_multi_moment(self):
+        # Test gen.multi with moment
+        # now that it's not a real Future
+        @gen.coroutine
+        def wait_a_moment():
+            result = yield gen.multi([gen.moment, gen.moment])
+            raise gen.Return(result)
+
+        loop = self.get_new_ioloop()
+        result = loop.run_sync(wait_a_moment)
+        self.assertEqual(result, [None, None])
+
+
 if __name__ == '__main__':
     unittest.main()
